<?php

namespace Drupal\proc\Form;

use Drupal\Component\Utility\Crypt;
use Drupal\Component\Utility\UrlHelper;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\InvokeCommand;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\File\FileSystem;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Link;
use Drupal\Core\Render\Markup;
use Drupal\Core\Session\AccountProxy;
use Drupal\Core\Url;
use Drupal\file\Entity\File;
use Drupal\file\FileRepository;
use Drupal\proc\Entity\Proc;
use Drupal\proc\ProcInterface;
use Drupal\proc\Traits\ProcRecipientTrait;
use Drupal\user\UserStorageInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;

/**
 * Encrypt content.
 */
class ProcEncryptForm extends FormBase {

  use ProcRecipientTrait;

  /**
   * The file system.
   *
   * @var \Drupal\Core\File\FileSystem
   */
  protected FileSystem $fileSystem;

  /**
   * The file repository.
   *
   * @var \Drupal\file\Entity\FileRepository
   */
  protected FileRepository $fileRepository;

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountProxy
   */
  protected AccountProxy $currentUser;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * The user storage.
   *
   * @var \Drupal\user\UserStorageInterface
   */
  protected UserStorageInterface $userStorage;

  /**
   * The configuration factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The current request.
   *
   * @var \Symfony\Component\HttpFoundation\Request
   */
  protected Request $request;

  /**
   * The label of the proc entity.
   *
   * @var string
   */
  protected String $label;

  /**
   * ProcEncryptForm form.
   *
   * @param \Drupal\Core\File\FileSystem $file_system
   *   The file system.
   * @param \Drupal\file\Entity\FileRepository $file_repository
   *   The file repository.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Session\AccountProxy $current_user
   *   The user account.
   * @param \Drupal\user\UserStorageInterface $user_storage
   *   The user storage.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory.
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The current request.
   */
  public function __construct(
      FileSystem $file_system,
      FileRepository $file_repository,
      EntityTypeManagerInterface $entity_type_manager,
      AccountProxy $current_user,
      UserStorageInterface $user_storage,
      ConfigFactoryInterface $config_factory,
      Request $request
    ) {
    $this->fileSystem = $file_system;
    $this->fileRepository = $file_repository;
    $this->entityTypeManager = $entity_type_manager;
    $this->currentUser = $current_user;
    $this->userStorage = $user_storage;
    $this->configFactory = $config_factory;
    $this->request = $request;
  }

  /**
   * {@inheritdoc}
   *
   * @param \Symfony\Component\DependencyInjection\ContainerInterface $container
   *   The Drupal service container.
   *
   * @return static
   */
  public static function create(ContainerInterface $container): static {
    return new static(
      $container->get('file_system'),
      $container->get('file.repository'),
      $container->get('entity_type.manager'),
      $container->get('current_user'),
      $container->get('entity_type.manager')->getStorage('user'),
      $container->get('config.factory'),
      $container->get('request_stack')->getCurrentRequest()
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'proc_encrypt_form';
  }

  /**
   * Dynamic title for Encrypt form.
   */
  public function getTitle() {
    $title = $this->t('Encrypt and upload');

    if (!$this->isFileInputMode()) {
      $title = $this->t('Encrypt');
    }

    return $title;
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $hidden_fields = [
      'cipher_text',
      'source_file_name',
      'source_file_size',
      'source_file_type',
      'source_file_last_change',
      'browser_fingerprint',
      'generation_timestamp',
      'generation_timespan',
      'host_page_url',
      'signed',
    ];

    foreach ($hidden_fields as $hidden_field) {
      $form[$hidden_field] = ['#type' => 'hidden'];
    }

    // Get query string:
    $query = $this->getRequest()->query->all();

    if (isset($query['host_page_url'])) {
      $form['host_page_url']['#value'] = $query['host_page_url'];
    }

    $field_input_mode = FALSE;
    // Get the input mode:
    if (isset($query['field-input-mode'])) {
      $field_input_mode = $query['field-input-mode'];
    }

    // File is the default input mode:
    if (!$field_input_mode) {
      $form['proc-file'] = [
        '#type' => 'file',
        '#description' => $this->t('Select a file for encryption.'),
        'data' => $query,
      ];
    }

    $field_type = '';
    if ((int) $field_input_mode === ProcInterface::INPUT_MODE_TEXTAREA) {
      $field_type = 'textarea';
    }
    if ((int) $field_input_mode === ProcInterface::INPUT_MODE_TEXTFIELD) {
      $field_type = 'textfield';
    }

    $form['proc-text-field'] = [
      '#type' => $field_type,
      '#description' => $this->t('Write a text for encryption.'),
    ];

    // Get current URL. If it has a query string standalone=FALSE, then we
    // do not use ajax submit.
    $proc_standalone_mode = FALSE;
    if (isset($query['proc_standalone_mode'])) {
      $proc_standalone_mode = $query['proc_standalone_mode'];
    }

    // If stand-alone mode is false, then this encryption
    // is triggered by a field widget and the dialog has to
    // be closed by ajax submit.
    if ($proc_standalone_mode == 'FALSE') {
      $form['#attached']['drupalSettings'] = [
        'proc' => [
          'proc_standalone_mode' => FALSE,
        ],
      ];

      $form['submit-proc'] = [
        '#type' => 'submit',
        '#value' => ($field_input_mode != ProcInterface::INPUT_MODE_FILE)
          ? $this->t('Encrypt')
          : $this->t('Encrypt and upload'),
        '#ajax' => [
          // Close dialog using ajax submit:
          'callback' => '::submitFormAjax',
          'event' => 'click',
        ],
      ];
    }
    else {
      $form['#attached']['drupalSettings'] = [
        'proc' => [
          'proc_standalone_mode' => TRUE,
        ],
      ];

      // Do not use ajax submit once this is not triggered by a field widget
      // and therefore there is no dialog to close:
      $form['actions']['submit-proc'] = [
        '#type' => 'submit',
        '#value' => ($field_input_mode != ProcInterface::INPUT_MODE_FILE)
          ? $this->t('Encrypt')
          : $this->t('Encrypt and upload'),
      ];
    }
    return $form;
  }

  /**
   * Submit handler for the ajax submit.
   */
  public function submitFormAjax(array &$form, FormStateInterface $form_state) {
    // Close the dialog:
    $response = new AjaxResponse();
    // Selector for the dialog close button.
    $selector = '.ui-dialog-titlebar-close';
    $method = 'click';
    $arguments = [];
    $response->addCommand(new InvokeCommand($selector, $method, $arguments));

    $file_name = (isset($form['proc-file']['#value']))
      ? $form['proc-file']['#value']["\0Symfony\\Component\\HttpFoundation\\File\\UploadedFile\0originalName"]
      : '';

    $proc_id = FALSE;
    // Get the latest proc ID authored by current user:
    $current_user_id = $this->currentUser->id();

    // If this proc was edited instead of created, use the known proc ID:
    $current_url = $form['submit-proc']['#attached']['drupalSettings']['ajaxTrustedUrl'];
    $current_url_string = key($current_url);
    $parsed_url = UrlHelper::parse($current_url_string);
    $edited_proc_id = FALSE;
    if (isset($parsed_url['query']['field-default-proc-id'])) {
      $edited_proc_id = $parsed_url['query']['field-default-proc-id'];
    }
    $proc_entity = FALSE;
    if ($edited_proc_id) {
      $proc_entity = Proc::load($edited_proc_id);
    }

    if ($proc_entity) {
      if ($proc_entity->getOwnerId() == $this->currentUser->id()) {
        $proc_id = $edited_proc_id;
      }
      else {
        $recipients = $proc_entity->get('field_recipients_set')->getValue();
        foreach ($recipients as $recipient) {
          if ($recipient['target_id'] == $this->currentUser->id()) {
            $proc_id = $edited_proc_id;
            break;
          }
        }
      }
    }

    // If there is no default proc ID, get the latest proc ID authored by current user:
    if (!$proc_id && !$edited_proc_id) {
      $query = $this->entityTypeManager->getStorage('proc')->getQuery();
      $query->accessCheck(TRUE);
      $query->condition('user_id', $current_user_id);
      $query->sort('id', 'DESC');
      $query->range(0, 1);
      $entity_ids = $query->execute();
      $proc_id = array_shift($entity_ids);
    }

    $proc_object_label = Proc::load($proc_id)->get('label')->getValue()[0]['value'];

    // If label was altered by hook proc label alter, use it:
    if (!empty($file_name) && $this->label != $file_name) {
      $file_name = $this->label;
    }

    if (!empty($file_name) && $file_name == $proc_object_label) {
      // Set the file name in the form:
      $field_name = $form_state->getCompleteForm()['#attached']['drupalSettings']['proc_field_name'];
      $delta = $form_state->getCompleteForm()['#attached']['drupalSettings']['proc_field_delta'];
      $selector = 'input[name="' . $field_name . '[' . $delta . '][target_id]"]';
      $file_name .= ' (' . $proc_id . ')';
    }
    if (empty($file_name)) {
      $selector = '#' . $form_state->getCompleteForm()['#attached']['drupalSettings']['proc_form_triggering_field'];
      // Text field encryption mode uses a hidden textfield as input mode and
      // therefore accepts only the cipher text ID:
      $file_name .= $proc_id;
    }

    $response->addCommand(new InvokeCommand($selector, 'val', [$file_name]));

    return $response;
  }

  /**
   * Implements form validation.
   *
   * The validateForm method is the default method called to validate input on
   * a form.
   *
   * @param array $form
   *   The render array of the currently built form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Object describing the current state of the form.
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    // @todo Add validation for the syntax of armored cipher text.
  }

  /**
   * Implements a form submit handler.
   *
   * The submitForm method is the default method called for any submit elements.
   *
   * @param array $form
   *   The render array of the currently built form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Object describing the current state of the form.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $recipients_set_ids = $form_state->get('storage');
    $current_url = $this->request->headers->get('referer');
    $json_content = $form_state->getValue('cipher_text');
    $meta = $this->extractMetaFields($form_state);
    $field_default_proc_id = $this->getFieldDefaultProcId($current_url);
    $proc = $this->createOrUpdateProcEntity($field_default_proc_id);

    $user_evaluation = $this->getRecipientUsers($recipients_set_ids);
    $recipient_users = $user_evaluation['recipient_users'];
    $rejected_users = $user_evaluation['rejected_users'];

    $files = $this->saveJsonFiles($json_content);
    $file_id = $files['file_id'];
    $json_fids = $files['json_fids'];

    if ($file_id) {
      $cipher = ['cipher_fid' => $file_id];
    }
    elseif (!empty($json_fids)) {
      $cipher = ['cipher_fid' => $json_fids];
    }

    $label = $meta['source_file_name'];
    // Implement hook proc label alter.
    \Drupal::moduleHandler()->alter('proc_label', $label, $meta, $form_state);

    $this->label = $label;

    $proc->set('armored', $cipher)
      ->set('meta', $meta)
      ->set('langcode', 'en')
      ->set('label', $label)
      ->set('type', 'cipher')
      ->set('field_recipients_set', $recipient_users)
      ->save();

    $proc_id = $proc->id();

    if (is_numeric($proc_id)) {
      $this->handleFileUsage($file_id, $proc_id, $json_fids);

      $link = $this->generateLink($proc_id);
      $this->logEncryptionCompletion($link);
      $this->handleFeedback($proc_id, $rejected_users);
    }
    else {
      $this->messenger()->addError($this->t('Error'));
    }
  }

  /**
   * Handles file usage.
   */
  private function handleFileUsage($file_id, $proc_id, $json_fids) {
    if ($file_id) {
      $this->addFileUsage($file_id, $proc_id);
    }

    if (!empty($json_fids)) {
      foreach ($json_fids as $json_fid) {
        $this->addFileUsage($json_fid, $proc_id);
      }
    }
  }

  /**
   * Adds file usage.
   */
  private function addFileUsage($file_id, $proc_id) {
    $file = File::load($file_id);
    $file_usage = \Drupal::service('file.usage');
    $file_usage->add($file, 'proc', 'file', $proc_id);
  }

  /**
   * Generates the decryption link.
   */
  private function generateLink($proc_id) {
    global $base_url;
    $url = Url::fromUri('internal:/proc/' . $proc_id);
    return Link::fromTextAndUrl($base_url . '/proc/' . $proc_id, $url)
      ->toString()
      ->getGeneratedLink();
  }

  /**
   * Logs encryption completion.
   */
  private function logEncryptionCompletion($link) {
    $this->getLogger('proc')->notice('Encryption is completed. Decryption link: %link', ['%link' => Markup::create($link)]);
  }

  /**
   * Handles feedback.
   */
  private function handleFeedback($proc_id, $rejected_users) {
    $config = $this->configFactory->get('proc.settings');

    if (!$config->get('proc-suppress-encryption-success-msg')) {
      $link = $this->generateLink($proc_id);

      if (!$config->get('proc-suppress-decrypt-link')) {
        $this->messenger()->addMessage($this->t('Encryption is completed. Decryption link: %link', ['%link' => Markup::create($link)]));
      }
      else {
        $this->messenger()->addMessage($this->t('Encryption is completed.'));
      }
    }

    if (!empty($rejected_users)) {
      $rejected_user_ids_csv = implode(', ', $rejected_users);
      $this->logRecipientImpact($proc_id, $rejected_user_ids_csv);
      $this->messenger()->addWarning($this->t('PGP key(ies) or recipient(s) not found on encrypted content (id: @proc_id). Rejected user ID(s): @rejected_users.', [
        '@proc_id' => $proc_id,
        '@rejected_users' => $rejected_user_ids_csv,
      ]));
    }
  }

  /**
   * Logs recipient impact.
   */
  private function logRecipientImpact($proc_id, $rejected_user_ids_csv) {
    $this->getLogger('proc')->warning('PGP key(ies) or recipient(s) not found on encrypted content (id: @proc_id). Rejected user ID(s): @rejected_users.', [
      '@proc_id' => $proc_id,
      '@rejected_users' => $rejected_user_ids_csv,
    ]);
  }

  /**
   * Extract meta fields from the form state.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Object describing the current state of the form.
   *
   * @return array
   *   Array containing meta field values.
   */
  private function extractMetaFields(FormStateInterface $form_state): array {
    return [
      'source_file_name' => $form_state->getValue('source_file_name'),
      'source_file_size' => $form_state->getValue('source_file_size'),
      'source_file_type' => $form_state->getValue('source_file_type'),
      'source_file_last_change' => $form_state->getValue('source_file_last_change'),
      'browser_fingerprint' => $form_state->getValue('browser_fingerprint'),
      'generation_timestamp' => $form_state->getValue('generation_timestamp'),
      'generation_timespan' => $form_state->getValue('generation_timespan'),
      'host_page_url' => $form_state->getValue('host_page_url'),
      'signed' => $form_state->getValue('signed'),
    ];
  }

  /**
   * Get the default proc ID from the current URL.
   *
   * @param string $current_url
   *   The current URL.
   *
   * @return int|false
   *   The default proc ID or false if not found.
   */
  private function getFieldDefaultProcId(string $current_url) {
    $parsed_url = UrlHelper::parse($current_url);
    $query_string = $this->request->get('field-default-proc-id');
    if (isset($parsed_url['query']['field-default-proc-id'])) {
      return (int) $parsed_url['query']['field-default-proc-id'];
    }
    if (isset($query_string)) {
      return (int) $query_string;
    }
    return FALSE;
  }

  /**
   * Create or update the Proc entity.
   *
   * @param int|false $field_default_proc_id
   *   The default proc ID.
   *
   * @return \Drupal\proc\Entity\Proc
   *   The Proc entity.
   */
  private function createOrUpdateProcEntity($field_default_proc_id) {
    $proc = Proc::create();
    if ($field_default_proc_id) {
      $proc_entity = Proc::load($field_default_proc_id);
      if ($proc_entity && $proc_entity->getOwnerId() == $this->currentUser->id()) {
        $proc = $proc_entity;
      }
    }
    return $proc;
  }

  /**
   * Get recipient users based on the selected strategy.
   *
   * @param array $recipients_set_ids
   *   Recipients set IDs.
   *
   * @return array
   *   Array of recipient users and/or rejected users.
   */
  private function getRecipientUsers(array $recipients_set_ids): array {
    $recipient_users = [];
    $rejected_users = [];

    switch ($this->configFactory->get('proc.settings')->get('proc-auto-filter-recipients')) {
      case ProcInterface::STRATEGY_RESTRICTIVE:
        foreach ($recipients_set_ids as $recipient_id) {
          $recipient_users[] = ['target_id' => $recipient_id];
        }
        break;

      case ProcInterface::STRATEGY_PERMISSIVE:
        foreach ($recipients_set_ids as $recipient_id) {
          if ($this->userHasKeyring($recipient_id)) {
            $recipient_users[] = ['target_id' => $recipient_id];
          }
          else {
            $rejected_users[] = $recipient_id;
          }
        }
        break;
    }

    return ['recipient_users' => $recipient_users, 'rejected_users' => $rejected_users];
  }

  /**
   * Save JSON content as separate files if needed.
   *
   * @param string $json_content
   *   JSON content to be saved.
   *
   * @return array
   *   Array of file IDs.
   */
  private function saveJsonFiles(string $json_content): array {
    $config = $this->configFactory->get('proc.settings');
    $enable_stream_wrapper = $config->get('proc-enable-stream-wrapper');
    $stream_wrapper = $config->get('proc-stream-wrapper');

    $json_fids = [];
    $file_id = FALSE;

    if ($enable_stream_wrapper === 1 && !empty($stream_wrapper)) {

      $block_size = $config->get('proc-file-block-size');

      $blocks_split_enabled = $config->get('proc-enable-block-size');

      $json_dest = $stream_wrapper;
      $this->createDirectory($json_dest);

      if ($json_content) {
        if ($blocks_split_enabled && !empty($block_size)) {
          $blocks_texts = $this->splitContentIntoBlocks($json_content, $block_size);
          foreach ($blocks_texts as $block_text) {
            $json_fids[] = $this->writeJsonBlock($json_dest, $block_text);
          }
        }
        else {
          $file_id = $this->writeJsonBlock($json_dest, $json_content);
        }
      }
    }
    return ['json_fids' => $json_fids, 'file_id' => $file_id];
  }

  /**
   * Create directory if it does not exist.
   *
   * @param string $directory
   *   The directory path.
   */
  private function createDirectory(string $directory): void {
    if (!is_dir($directory)) {
      $this->fileSystem->mkdir($directory, NULL, TRUE);
    }
  }

  /**
   * Split content into blocks based on block size.
   *
   * @param string $json_content
   *   JSON content to be split.
   * @param int $block_size
   *   Size of each block.
   *
   * @return array
   *   Array of content blocks.
   */
  private function splitContentIntoBlocks(string $json_content, int $block_size): array {
    $lines = explode("\n", $json_content);
    $content_lines_number = count($lines);
    $lines_size_ratio = $content_lines_number / $block_size;
    $blocks = intval($lines_size_ratio);
    $remaining = $content_lines_number % $block_size;

    if ($remaining > 0) {
      $blocks++;
    }

    $blocks_lines = [];
    $content_line_index = 0;
    $blocks_index = 0;

    while ($blocks_index < $blocks) {
      $line_in_block_index = 0;
      while ($line_in_block_index < $block_size) {
        if (isset($lines[$content_line_index])) {
          $blocks_lines[$blocks_index][] = $lines[$content_line_index];
        }
        $content_line_index++;
        $line_in_block_index++;
      }
      $blocks_index++;
    }
    $blocks_texts = [];
    foreach ($blocks_lines as $block_index => $block_lines) {
      foreach ($block_lines as $block_line) {
        $blocks_texts[$block_index] = $blocks_texts[$block_index] . "\n" . $block_line;
      }
    }

    return $blocks_texts;
  }

  /**
   * Write JSON block to a file and return the file ID.
   *
   * @param string $json_dest
   *   Destination directory for JSON files.
   * @param string $block_text
   *   JSON block content.
   *
   * @return int
   *   File ID of the written JSON block.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  private function writeJsonBlock(string $json_dest, string $block_text): int {
    $fragment = Crypt::hashBase64(Crypt::randomBytesBase64(32));
    $json_filename = $fragment . '.json';
    $jsonFid = $this->fileRepository->writeData(
      $block_text,
      "$json_dest/$json_filename",
      FileSystemInterface::EXISTS_REPLACE
    );

    return $jsonFid->id();
  }

  /**
   * Validate Proc field input mode.
   */
  private function isFileInputMode() {
    $input_mode = ($this->request->query->has('field-input-mode'))
      ? $this->request->query->get('field-input-mode')
      : NULL;
    if (!isset($input_mode)) {
      return FALSE;
    }

    return ((int) $input_mode === ProcInterface::INPUT_MODE_FILE);
  }

}
